#!BPY

"""
Name: 'BVH 4 BLENDER'
Blender: 234
Group: 'Wizards'
Tooltip: 'Create Armature on BVH Empties'
"""

######################################################
# BVH4BLEND.py v1.2
#
# Script for generating armature on BVH empties
#
#####################################################
# This script generates an armature and make bones
# follow empties created by Blender BVH import 
# How to use it :
# - Import a bvh in Blender (Fie->Import->BVH)
# - Launch this script (Alt-P)
# - Set up variables : 
#   "startframe"  the first frame of your anim
#   "endframe"  the last frame of your anim
#   "hipbonename" the name of the main bone
# - Press "Create Armature"
######################################################
# Author : Jean-Baptiste PERIN (jb_perin@yahoo.fr)
######################################################



import Blender
from Blender import Mathutils
import math

dicEmptiesRestMatrix= {}
dicEmptiesInvRestMatrix= {}
dicBoneRestMatrix= {}
dicBone={}
dicEmptyChild={}
dicBoneRestInvEmpRest={}
dicEmpRestInvBoneRest={}

########################################################################
#
#          UTILITY FUNCTIONS FOR HANDLING BONES AND EMPTIES
#
########################################################################


#########
# Cette fonction renvoie la liste des empties attaches a root
# in  : 
# out : emp_list (List of Object) la liste des objets de type "Empty"
#########
def getTree(emp_list, root):
	empties=getAllEmpties()
	chlds = getChildren(root, empties)
	dicEmptyChild[root.getName()]=chlds
	for ch in chlds:
		emp_list.append(ch)
		getTree(emp_list,ch)

#########
# Cette fonction renvoie la liste des empties attaches a root
# in  : 
# out : emp_list (List of Object) la liste des objets de type "Empty"
#########
def getEmpties():
	global hipbonename
	emp_list = []
	root = Blender.Object.Get(hipbonename)
	emp_list.append(root)
	getTree(emp_list, root)
	return emp_list

#########
# Cette fonction renvoie la liste des empties
# in  : 
# out : emp_list (List of Object) la liste des objets de type "Empty"
#########
def getAllEmpties():
	emp_list = []
	objs = Blender.Object.Get()
	for o in objs:
		if o.getType()=="Empty":
			emp_list.append(o)
	return emp_list

#########
# Cette fonction renvoie la liste des empties
# in  : 
# out : emp_list (List of Object) la liste des objets de type "Empty"
#########
def getEmpty(name):
	p = None
	objs = Blender.Object.Get()
	for o in objs:
		if o.getType()=="Empty" and o.getName()==name:
			p = o
	return p

def getChild(emp, emp_list):
	return dicEmptyChild[emp.getName()]


#########
# Cette fonction fournit la liste des enfants d'un empty
# in  : emp (Object) un empty
#       emp_list (List of Object) la liste des empties
# out : children (List of Object) la liste des empties enfants de 'empty'
#########
def getChildren(emp, emp_list):
	children = []
	root_emp = getRootEmpties(emp_list)
	for em in emp_list:
		if (em.getName() != emp.getName()) and (em not in root_emp):
			if (em.getParent().getName() == emp.getName()):
				children.append(em)
	return children



#########
# Cette fonction renvoie la liste des empties n'ayant pas de parent
# in  : emp_list (List) une liste d'empties
# out : root (List) le (ou les) empty n'ayant pas de parent
#########
def getRootEmpties(emp_list):
	root = []
	for em in emp_list:
		if em.getParent() == None:
			root.append(em)
	return root


#########
# Cette fonction renvoie le bone de nom 'name' dans l'armature 'armature'
# in  : armature (Armature) l'armature dans laquelle cherchait le bone
#       name (String) le nom de l'os a chercher
# out : p (Bone) 
#########
#def getBone(armature, name):
#	return (dicBone[name])
	#p = None
	#bones = armature.getBones()
	#for i in bones:
	#	if i.getName() == name:
	#		p = i
	#		break
	#return p




########################################################################
#
# FUNCTIONS FOR COMPUTING POSITION AND ROTATION OF BONES 
#
########################################################################

def computeRootQuat2(empty, bone):

	M1=dicBoneRestInvEmpRest[bone.getName()].rotationPart()
	M2=dicEmpRestInvBoneRest[bone.getName()].rotationPart()
	emprot = empty.getMatrix('worldspace').rotationPart()
	emprot.transpose()
	mat = M1*emprot*M2
	mat.transpose()
	return (mat.toQuat())

	#emprest = dicEmptiesRestMatrix[empty.getName()].rotationPart()
	#invemprest= dicEmptiesRestMatrix[empty.getName()].rotationPart()
	##invemprest= emprest
	##invemprest.invert()
	##invemprest= dicEmptiesInvRestMatrix[empty.getName()].rotationPart()
	#emprot = empty.getMatrix('worldspace').rotationPart()
	#bonerest = dicBoneRestMatrix[bone.getName()].rotationPart()
	#invbonerest = dicBoneRestMatrix[bone.getName()].rotationPart()
	#invbonerest.invert()
	#T2=emprot*invemprest
	#T2.transpose()
	#mat = bonerest*invemprest*T2*emprest*invbonerest
	#mat.transpose()
	#return (mat.toQuat())




#########
# Cette fonction 
# in  : 
# out : 
#########
def computeRootPos(empty, bone):
	vec = empty.getMatrix('worldspace').translationPart() - dicBoneRestMatrix[bone.getName()].translationPart()
	mat = dicBoneRestMatrix[bone.getName()].rotationPart()
	vec2 =  Mathutils.MatMultVec (mat, vec)
	return vec2


########################################################################
#
#            FUNCTIONS FOR CREATING AND MOVING ARMATURES
#
########################################################################

#########
# Cette fonction 
# in  : 
# out : 
#########
def createBone (armature, empty, bone, empties):
	children = getChildren(empty, empties)
	for ch in children:
		if len(children) >= 2:
			bonename = empty.getName()[1:len(empty.getName())]+'_'+ch.getName()[1:len(ch.getName())]
		else :
			bonename = empty.getName()[1:len(empty.getName())]
		b=Blender.Armature.Bone.New(bonename)
		b.setHead(empty.getMatrix('worldspace').translationPart())
		b.setTail(ch.getMatrix('worldspace').translationPart())
		#b.setParent(bone)
		matrice = empty.getMatrix('worldspace')
		invmatrice = empty.getMatrix('worldspace')
		invmatrice.invert()
		invmatricet=empty.getMatrix('worldspace')
		invmatricet.invert()
		invmatricet.transpose()
		dicEmptiesRestMatrix[empty.getName()] = matrice
		dicEmptiesInvRestMatrix[empty.getName()] = invmatrice
		armature.addBone(b)
		invbonerest=b.getRestMatrix()
		invbonerest.invert()
		dicBoneRestMatrix[b.getName()] = b.getRestMatrix()
		dicBoneRestInvEmpRest[b.getName()]=b.getRestMatrix()*invmatrice*invmatricet
		dicEmpRestInvBoneRest[b.getName()]=matrice*invbonerest
		dicBone[b.getName()]=b
		createBone(armature, ch, b, empties)

	

#########
# Cette fonction fabrique une arma
# in  : 
# out : 
#########
def createArmature (rootEmpty, empties):
	armData = Blender.Armature.New('monArmature')
	children = getChildren(rootEmpty, empties)
	for ch in children:
		b=Blender.Armature.Bone.New(rootEmpty.getName()[1:len(rootEmpty.getName())] + ch.getName()[1:len(ch.getName())])
		b.setHead(rootEmpty.getMatrix('worldspace').translationPart())
		b.setTail(ch.getMatrix('worldspace').translationPart())
		armData.addBone(b)
		matrice = ch.getMatrix('worldspace')
		invmatrice = ch.getMatrix('worldspace')
		invmatrice.invert()
		invmatricet=ch.getMatrix('worldspace')
		invmatricet.invert()
		invmatricet.transpose()
		dicEmptiesRestMatrix[rootEmpty.getName()] = matrice
		dicEmptiesInvRestMatrix[rootEmpty.getName()] = invmatrice
		invbonerest=b.getRestMatrix()
		invbonerest.invert()
		dicBoneRestMatrix[b.getName()] = b.getRestMatrix()
		dicBoneRestInvEmpRest[b.getName()]=b.getRestMatrix()*invmatrice*invmatricet
		dicEmpRestInvBoneRest[b.getName()]=matrice*invbonerest
		dicBone[b.getName()]=b
		createBone(armData, ch, b, empties)
	return armData

#########
# Cette fonction 
# in  : 
# out : 
#########
def moveBones(armature, empty, empties):
	children = dicEmptyChild[empty.getName()]
	for ch in children:
		if len(children) >= 2:
			bonename = empty.getName()[1:len(empty.getName())]+'_'+ch.getName()[1:len(ch.getName())]
		else :
			bonename = empty.getName()[1:len(empty.getName())]
		bone = dicBone[bonename]
		bone.setLoc(computeRootPos(empty,bone))
		bone.setQuat(computeRootQuat2(empty,bone))
		chch = dicEmptyChild[ch.getName()]
		if len(chch) >= 1:
			moveBones(armature, ch, empties)


#########
# Cette fonction 
# in  : 
# out : 
#########
def moveArmature (armature, empties):
	root = Blender.Object.Get(hipbonename)
	children = dicEmptyChild[hipbonename]
	for ch in children:
		b=dicBone[hipbonename[1:len(hipbonename)] + ch.getName()[1:len(ch.getName())]]
		b.setLoc(computeRootPos(root, b))
		b.setQuat(computeRootQuat2(root, b))
		moveBones(armature, ch, empties)




########################################################################
#
#                  MAIN PROGRAM
#
########################################################################

def Main():

	global endframe, startframe,hipbonename


	print "*****START*****"

	Blender.Set("curframe",startframe)

	##-----------
	## Positionnement des empties
	##-----------
	em0 = Blender.Object.Get(hipbonename)

	Blender.Redraw()



	##-----------
	## Creation de l'armature et des os
	##-----------

	lesEmpties = getEmpties()
	#print dicEmptyChild	
	armData = createArmature(em0, lesEmpties)
	#armData.drawAxes(1)
	#armData.drawNames(1)

	##-----------
	## Ajout de l'armature a la scene
	##-----------
	armObj=Blender.Object.New('Armature', 'OBArmature')

	armObj.link(armData)
	scn = Blender.Scene.getCurrent()
	scn.link (armObj)
	Blender.Redraw()

	##-----------
	## Enregistrement de la position  de l'armature
	##-----------

	bones = armData.getBones()
	for bo in bones:
		bo.setPose([Blender.Armature.Bone.ROT, Blender.Armature.Bone.LOC]) 

	Blender.Redraw()


	while endframe >= Blender.Get("curframe"):
		##-----------
		## Passage a la frame suivante
		##-----------
		num_frame = Blender.Get("curframe")
		print num_frame
		Blender.Set("curframe", num_frame+1)

		##-----------
		## Positionnement des os
		##-----------

		moveArmature(armData, lesEmpties)


		##-----------
		## Enregistrement de la position  de l'armature
		##-----------

		for bo in bones:
			bo.setPose([Blender.Armature.Bone.ROT, Blender.Armature.Bone.LOC]) 

	Blender.Set("curframe",startframe)
	Blender.Redraw()

	print "*****END*****"


########################################################################
#
#            GUI FUNCTIONS AND VARIABLES
#
########################################################################

EFrame = Blender.Draw.Create(0)
SFrame2 = Blender.Draw.Create(0)
HBName = Blender.Draw.Create(0)
Msg = ' '

def event (evt, val):
	if evt == Blender.Draw.ESCKEY:
		Blender.Draw.Exit()
	return

def button_event(evt):
	global EFrame, SFrame2, HBName, Msg
	global endframe, startframe, hipbonename
	if evt==1:
		startframe = eval(SFrame2.val)
		endframe =  eval(EFrame.val)
		hipbonename = HBName.val
		if startframe>=endframe:
			Msg = 'Start frame must be lower than End frame'
		else:
			ob = getEmpty(hipbonename)
			if (ob!=None): 
				if ob.getParent()!=None:
					Msg = 'Empty '+hipbonename+ ' is not a root bone.'
				else:  
					Blender.Draw.Exit()
					Main()
			else:
				Msg = 'Empty '+ hipbonename+ ' not found'
				
		Blender.Draw.Redraw(1)

def GUI():
	global EFrame, SFrame2, HBName, Msg
	Blender.BGL.glClearColor(0,0,1,1)
	Blender.BGL.glClear(Blender.BGL.GL_COLOR_BUFFER_BIT)
	Blender.BGL.glColor3f(1,1,1)
	Blender.BGL.glRasterPos2i(20,150)
	Blender.Draw.Text ("BVH 4 BLENDER v1.2 by Jean-Baptiste PERIN", 'normal')
	EFrame = Blender.Draw.String("Endframe: ", -1, 20, 100, 250, 20, str(Blender.Get("endframe")), 100)
	HBName = Blender.Draw.String("HipBoneName: ", -1, 20, 75, 250, 20, '_Hips', 100)
	SFrame2 = Blender.Draw.String("Startframe: ", -1, 20, 125, 250, 20, str(Blender.Get("staframe")), 100)
	Blender.Draw.Toggle("Create Armature", 1, 20, 10, 100, 20, 0, "Create Armature")
	Blender.BGL.glRasterPos2i(20,40)
	Blender.Draw.Text (Msg, 'normal')


Blender.Draw.Register(GUI, event, button_event)