#!BPY

"""
Name: 'TRACKBONE'
Blender: 234
Group: 'Wizards'
Tooltip: 'Make an empty track a bone'
"""

######################################################
# TRACKBONE.py v1.0
#
# Script for creating an empty an make it follow a bone
#
#####################################################
# This script creates an empty which follows a 
# specified bone into a selected armature 
# How to use it :
# - Select an armature
# - Launch this script (Alt-P)
# - Set up variables : 
#   "bonename" the name of bone you want to track
#   "startframe"  the first frame of your anim
#   "endframe"  the last frame of your anim
# - Press "Create Track"
######################################################
# Author : Jean-Baptiste PERIN (jb_perin@yahoo.fr)
######################################################



import Blender
from Blender import *


scn = Blender.Scene.getCurrent()

def names(ob): return ob.getName()


#########
# Cette fonction renvoie le bone de nom 'name' dans l'armature 'armature'
# in  : armature (Armature) l'armature dans laquelle cherchait le bone
#       name (String) le nom de l'os a chercher
# out : p (Bone) 
#########
def getBone(armature, name):
	p = None
	bones = armature.getBones()
	for i in bones:
		if i.getName() == name:
			p = i
			break
	return p

#########
# Cette fonction renvoie un empty 
# in  : objname : le nom de l'empty recherche
# out : myobj : l'empty cree ou retrouve
#########
def GetOrCreateEmpty(objname):
	objects = Blender.Object.Get()
	if objname in map(names,objects):
		myobj = Blender.Object.Get(objname)
		print objname+' exists'
	else:
		myobj = Blender.Object.New('Empty',objname)

		scn.link(myobj)	
		print objname+' was created'
	return myobj


#########
# Cette fonction calcul la position dans le monde de l'os bone 
# in  : armobj : l'armature contenant l'os bone
#		 bone  : l'os dont on cherche la position dans le monde
# out : empworldpos : vecteur position de l'os bone dans le monde
#########
def computeEmptyPos(armobj, bone ):
	bonerelpos = bone.getRestMatrix('worldspace').translationPart()
	rotarm = armobj.getMatrix('worldspace').rotationPart()
	rotarm.transpose()
	armabspos = armobj.getMatrix('worldspace').translationPart()
	boneabspos=Mathutils.MatMultVec (rotarm, bonerelpos)
	empworldpos = boneabspos+armabspos
	return empworldpos


########################################################################
#
#                  MAIN PROGRAM
#
########################################################################

def Main():

	global endframe, startframe,bonename


	print "*****START*****"

	Blender.Set("curframe",startframe)

	##-----------
	## Recuperation de l'armature et de l'os cible
	##-----------
	objArm = Blender.Object.GetSelected()[0]
	arm = objArm.getData()
	lebone = getBone(arm, bonename)
	
	##-----------
	## Creation de l'empty
	##-----------
	
	emptyCible=GetOrCreateEmpty('Track'+objArm.getName())

	
	emptyCible.clearIpo()

	ipoEmpty = Ipo.New('Object','ipoEmp')

	emptyCible.setIpo(ipoEmpty)

	emptyloc = computeEmptyPos(objArm, lebone)

	emptyCiblePosX=ipoEmpty.addCurve('LocX')
	emptyCiblePosY=ipoEmpty.addCurve('LocY')
	emptyCiblePosZ=ipoEmpty.addCurve('LocZ')


	emptyCiblePosX.addBezier((Blender.Get("curframe"), emptyloc[0]))
	emptyCiblePosY.addBezier((Blender.Get("curframe"), emptyloc[1]))
	emptyCiblePosZ.addBezier((Blender.Get("curframe"), emptyloc[2]))



	while endframe >= Blender.Get("curframe"):

		##-----------
		## Passage a la frame suivante
		##-----------
		num_frame = Blender.Get("curframe")
		print num_frame
		Blender.Set("curframe", num_frame+1)

		##-----------
		## Enregistrement de la position de l'empty
		##-----------
		emptyloc = computeEmptyPos(objArm, lebone)

		emptyCiblePosX=ipoEmpty.addCurve('LocX')
		emptyCiblePosY=ipoEmpty.addCurve('LocY')
		emptyCiblePosZ=ipoEmpty.addCurve('LocZ')


		emptyCiblePosX.addBezier((Blender.Get("curframe"), emptyloc[0]))
		emptyCiblePosY.addBezier((Blender.Get("curframe"), emptyloc[1]))
		emptyCiblePosZ.addBezier((Blender.Get("curframe"), emptyloc[2]))


	emptyCiblePosX.Recalc()
	emptyCiblePosY.Recalc()
	emptyCiblePosZ.Recalc()

	Blender.Set("curframe",startframe)
	Blender.Redraw()

	print "*****END*****"




########################################################################
#
#            GUI FUNCTIONS AND VARIABLES
#
########################################################################

EFrame = Blender.Draw.Create(0)
SFrame2 = Blender.Draw.Create(0)
HBName = Blender.Draw.Create(0)
Msg = ' '

def event (evt, val):
	if evt == Blender.Draw.ESCKEY:
		Blender.Draw.Exit()
	return

def button_event(evt):
	global EFrame, SFrame2, HBName, Msg
	global endframe, startframe, bonename
	if evt==1:
		startframe = eval(SFrame2.val)
		endframe =  eval(EFrame.val)
		bonename = HBName.val
		if startframe>=endframe:
			Msg = 'Start frame must be lower than End frame'
		else:
			obs = Object.GetSelected()
			if (len(obs)!=1): 
				Msg = 'One and only one armature can be selected ...'
			else:
				if obs[0].getType()!='Armature':
					Msg = 'The selected object must be an armature'
				else: 
					bo = getBone(obs[0].getData(), bonename)
					if (bo!=None): 
						Blender.Draw.Exit()
						Main()
					else:
						Msg = 'Unable to find bone '+ bonename+ ' in selected armature'
				
		Blender.Draw.Redraw(1)

def GUI():
	global EFrame, SFrame2, HBName,  Msg
	Blender.BGL.glClearColor(0,0,1,1)
	Blender.BGL.glClear(Blender.BGL.GL_COLOR_BUFFER_BIT)
	Blender.BGL.glColor3f(1,1,1)
	Blender.BGL.glRasterPos2i(20,260)
	Blender.Draw.Text ("TRACK BONE v1.0 by Jean-Baptiste PERIN", 'normal')

	Blender.BGL.glRasterPos2i(20,220)
	Blender.Draw.Text ("- Select the armature the bone is in", 'normal')
	Blender.BGL.glRasterPos2i(20,200)
	Blender.Draw.Text ("- Set up variables ", 'normal')
	Blender.BGL.glRasterPos2i(20,180)
	Blender.Draw.Text ("- Press button 'Create Track'", 'normal')

	EFrame = Blender.Draw.String("Endframe: ", -1, 20, 100, 250, 20, str(Blender.Get("endframe")), 100)
	HBName = Blender.Draw.String("BoneName: ", -1, 20, 75, 250, 20, 'HipsChest', 100)
	SFrame2 = Blender.Draw.String("Startframe: ", -1, 20, 125, 250, 20, str(Blender.Get("staframe")), 100)
	Blender.Draw.Toggle("Create Track", 1, 20, 10, 100, 20, 0, "Create Track")
	Blender.BGL.glRasterPos2i(20,40)
	Blender.Draw.Text (Msg, 'normal')


Blender.Draw.Register(GUI, event, button_event)
